#!/usr/bin/perl
# Test if decayed sum grows significantly large
$usage = <<'EOF' ;
usage: decay.pl [-h] [-n <iterations>] [-l <decay_rate>] [-r <randomness>] [-I <input> [-t <input type>]
	-h : help
	<iterations> : integer.
	<decay_rate> : real number.
	<randomness> : real number [0:1).
	<input> : real number.
	<input type> : 0 = random graded
		       1 = alternating step
		       3 = sinusoidal
EOF

require "getopts.pl";
&Getopts('hl:n:r:I:t:');

if ($opt_h) {
	print $usage;
	exit(0);
}

# Default values
$opt_n = ($opt_n)?($opt_n):1000;
$opt_l = ($opt_l)?($opt_l):0.5;
$opt_r = ($opt_r)?($opt_r):0.0;
$opt_I = ($opt_I)?($opt_I):1.0;
$opt_t = ($opt_t)?($opt_t):0;

$sum = 0.0;

$asymptote = 1.0/(1.0-exp(-$opt_l)) * $opt_I;
$timeconst= $asymptote * (1.0-exp(-1.0));

@inputs = (
	# Random Graded Input
	'$I = $opt_I*($opt_r*rand(10)/10.0 + (1-$opt_r))',
	# Alternating Step Input: 
	'$I = (($i/40)%2==0)?1.0:0.0;',
	# Sinusoidal : 
	'$I = sin($i*3.141592/180.0*10.0)');

for ($i=0; $i<$opt_n; ++$i) {

	$I = eval($inputs[$opt_t]);
	print "$i ".$sum." $asymptote $timeconst $I\n";
	$sum = $sum * exp(-$opt_l) + $I;

}
