//----------------------------------------------------------------------------
// 
//	file: ppmio.h
// 
//	function: ppm input output routines 
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Sat Nov  7 03:55:17 CST 1998
// 
//	$Revision: 1.6 $ $Date: 2004/06/03 23:40:34 $
//	$Log: ppmio.h,v $
//	Revision 1.6  2004/06/03 23:40:34  yschoe
//	*** empty log message ***
//
//	Revision 1.4  2001/11/30 20:24:15  yschoe
//	pglissom v1.0 release
//
//	Revision 1.3  2000/02/08 11:25:21  yschoe
//	*** empty log message ***
//
//	Revision 1.2  1999/02/09 07:42:25  yschoe
//	hsv mode added
//
//	Revision 1.1  1998/11/17 12:51:24  yschoe
//	Initial revision
//
//
//----------------------------------------------------------------------------
#ifndef _PPMIO_H_
#define _PPMIO_H_

#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "util.h"
#include "real.h"
#include "hsvrgb.h"

enum PPM_COLOR_MODE {PPM_HSV,PPM_RGB,PPM_GRAY,PPM_GRAY2,PPM_HSV2,PPM_HSV3,PPM_TRUE_RGB};
enum PPM_TYPE {PPM_REAL,PPM_INT};
enum PPM_FLIP_TYPE {PPM_FLIP,PPM_NOFLIP};

typedef struct color_t   {
        int     red;
        int     green;
        int     blue;
} ctab_type;

class ppmio {

private: 


  REAL **mat; // hue: this is a ptr to the matrix to be dumped : don't alloc it
  REAL **mat2; // saturation or value
  REAL **mat3; // value
  REAL **rmat;
  REAL **gmat;
  REAL **bmat;
  REAL **hmat;
  REAL **smat;
  REAL **vmat;
  
  int	cols; // matrix cols
  int	rows; // matrix rows
  PPM_TYPE type;

  REAL value_mult; // hsv value multiplication factor
  REAL saturation_mult; // hsv value multiplication factor

  // fill in color table : color_tab[index] --> RGB value
  ctab_type *color_tab; 
  ctab_type *gray_tab; 
  ctab_type *hsv_tab; 
  ctab_type *hsv_tab2; 

  int 	colors;	// number of colors to be used : default = 256 (8bit)

  void init_color_tab(int colors);

public:

  ppmio(int **mat, int cols, int rows, int colors = 256);
  ppmio(REAL **mat, int cols, int rows, int colors = 256);
  ppmio(REAL **rmat, REAL **gmat, REAL **bmat, int cols, int rows, int colors = 256);
  ppmio(char t, REAL **rmat, REAL **gmat, REAL **bmat, int cols, int rows, int colors = 256);
  ppmio(REAL **mat, REAL **mat, int cols, int rows, int colors, 
		REAL mult);
  ppmio(REAL **mat, REAL **mat, REAL **mat, int cols, int rows, int colors, REAL mult, REAL mult2); 
  ppmio(REAL *vec, int cols, int rows, int colors = 256);

  ~ppmio();

  // Write to file "filename" with min and max scaled to given value.
  // If 0.0 is given for both min and max, the actual min and max values
  // are calculated and used.
  void ppmwrite(char *filename, REAL min=0.0, REAL max=0.0, 
		PPM_COLOR_MODE m=PPM_RGB, PPM_FLIP_TYPE flip_flag=PPM_FLIP);

  void ppmansi(FILE *fp=stdout, REAL min=0.0, REAL max=0.0, 
		PPM_COLOR_MODE m=PPM_RGB);

  // Read from a file.
  void ppmread(char *filename);

  //Read from a file(color).
  void ppmreadrgb(char *filename);
  
 //Read from a file(color - hsv).
  void ppmreadhsv(char *filename);

  // Read ppm header from file
  int  ppmheader(FILE *fp, int *ppmcols, int *ppmrows, int *ppmcolors);

  // print colormap in octave format
  void ppmcolormap(char *filename, PPM_COLOR_MODE m=PPM_RGB);

  void RGBtoHSV(unsigned char r,unsigned char g,unsigned char b, REAL *h, REAL *s, REAL *v );

  void ppmio::hsv_to_rgb();
  void ppmio::hsv_to_rgb_convert(unsigned char *r,unsigned char *g,unsigned char *b,REAL h,REAL s,REAL v);

  REAL Minval(REAL r, REAL g, REAL b);
  REAL Maxval(REAL r, REAL g, REAL b);

};

#endif
