//----------------------------------------------------------------------------
// 
//	file: lissom.cc
// 
//	function: defines and implements LISSOM 
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Wed Aug  5 18:04:55 CDT 1998
// 
//	$Revision: 1.1 $ $Date: 1998/09/04 08:42:31 $
//	$Log: lissom-text.cc,v $
//	Revision 1.1  1998/09/04 08:42:31  yschoe
//	got rid of GUI from lissom.cc
//
//	Revision 1.4  1998/08/26 02:37:20  yschoe
//	changed run() for more interactiveness.
//
//	Revision 1.3  1998/08/17 19:24:09  yschoe
//	map size N = 21 : works well with
//	lissom.conf 1.1 , netlayer v1.12
//
//	Revision 1.2  1998/08/17 16:30:44  yschoe
//	added more logs with netlog.
//
//	Revision 1.1  1998/08/17 00:15:51  yschoe
//	netlayer v1.12, 12x12 map, about 5000 iterations, connection death,etc. works file. use with lissom.conf (same version)
//
//
//----------------------------------------------------------------------------
#include <iostream.h>
#include <nnview/netlayer.h>
#include <nnview/simulog.h>
#include <nnview/real.h>
#include <nnview/coord.h>

int             steps;
int 		iters = 101;
int		disp_rate = 100;
int		iters_per_run = 100;
int 		t;
simulog		*netlog;

netlayer	*input;
netlayer	*map;

coord		**cog;
coord		**ideal;


int N = 15;
int M = 3;
REAL randomness=0.4;

//----------------------------------------------------------------------------
//      setup() 
//        - bplayer : setup network for simulation
//----------------------------------------------------------------------------
void setup(char *configfile) {

  netlog = new simulog("lissom.cc","lissom.log");

  // input layer
  cout << "Initializing input layer.\n";
  input = new netlayer("input",1,M,0,0,0);
  
  // map layer
  cout << "Initializing map layer.\n";
  map   = new netlayer("map",N,N,3,(int)(N*1.0),1,input,1,M,M);

  // set parameter values
  //   read from file.
  FILE *conffp = fopen(configfile,"r");
  char buff[100];

  // read parameters
  int count = 0;
  while (!feof(conffp)) {
    	fgets(buff,100,conffp);
	if (feof(conffp)) { break; }
        map->set_parameter(buff);
	count ++;
  }
  cout << "lissom:setup() " << count << " commands processed\n";
  fclose(conffp);

  // init weights
  map->init_weights(0.0,1.0);
  map->normalize_weights();

  // normalize weight to cartesian coord
  REAL cosproduct;
  REAL w[2];
  for (int i=0; i<map->cols; ++i) {
    for (int j=0; j<map->rows; ++j) {
      cosproduct=1.0;
      w[0] = ((1.0-randomness)*i/(REAL)(map->cols-1) 
	   + randomness*drand48()); 
      w[1] = ((1.0-randomness)*j/(REAL)(map->rows-1)
	   + randomness*drand48()); 
      map->wgt[i][j].aff_ws[0][0][0] = cos(w[0])*cos(w[1]);
      map->wgt[i][j].aff_ws[0][0][1] = sin(w[0])*cos(w[1]);
      map->wgt[i][j].aff_ws[0][0][2] = sin(w[1]);
    }
  }
  

  // set up ideal grid
  // ideal = NULL;
  ideal = new coord* [N];
  for (int i=0; i<N; ++i) 
	ideal[i] = new coord [N];

  for (int i=0; i<map->cols; ++i) {
    for (int j=0; j<map->rows; ++j) {
       w[0] = (REAL)i/((REAL)map->cols-1);
       w[1] = (REAL)j/((REAL)map->cols-1);
       ideal[i][j].x = w[0];
       ideal[i][j].y = w[1];
    }
  }

  // set up COG (center of gravity) map
  cog = new coord* [N];
  for (int i=0; i<N; ++i) 
	cog[i] = new coord [N];

}


//----------------------------------------------------------------------------
//	calc_cog() 
//	  - calculate center of gravity coordinates
//----------------------------------------------------------------------------
void calc_cog(coord **cog, netlayer *map, int cur_inp_layer){

  for (int x=0; x<map->cols; ++x) {

     for (int y=0; y<map->rows; ++y) {

	// calculate sums for center of gravity 
        REAL sum = 0.0;
        REAL x_sum = 0.0;
        REAL y_sum = 0.0;
	/*
	for (int i=map->rf_ranges[cur_inp_layer][x][y].lowx;
		 i<=map->rf_ranges[cur_inp_layer][x][y].highx; ++i) {
	   for (int j=map->rf_ranges[cur_inp_layer][x][y].lowy;
		    j<=map->rf_ranges[cur_inp_layer][x][y].highy; ++j) {
		int rf_x = i-map->rf_ranges[cur_inp_layer][x][y].lowx;
		int rf_y = j-map->rf_ranges[cur_inp_layer][x][y].lowy;
		REAL wgt = map->wgt[x][y].aff_ws[cur_inp_layer][rf_x][rf_y];
		sum += wgt;
		x_sum += i*wgt;
		y_sum += j*wgt;
	   }
      	}	
	*/
		
	// get the 3D weights
        REAL x_wgt = map->wgt[x][y].aff_ws[cur_inp_layer][0][0]; //x_sum/sum;
        REAL y_wgt = map->wgt[x][y].aff_ws[cur_inp_layer][0][1]; //y_sum/sum;
        REAL z_wgt = map->wgt[x][y].aff_ws[cur_inp_layer][0][2]; //y_sum/sum;

	// now save the actual c.o.g. after transforming back to 2D
 	cog[x][y].x = asin(y_wgt/cos(asin(z_wgt)));
	cog[x][y].y = asin(z_wgt); ///cos(asin(x_wgt)));

     }

  }

}

//----------------------------------------------------------------------------
//      run() 
//        - bplayer : run a certain number of backprop learning
//----------------------------------------------------------------------------
void run () {

  char str[100];
  int i;

  // for the number of given steps
  for (i=t; i<=t+iters_per_run; ++i) {

    // set input and normalize
    REAL sum=0.0;
    REAL x_inp = drand48();
    REAL y_inp = drand48();
    input->output[0][0] = cos(x_inp)*cos(y_inp);
    input->output[0][1] = sin(x_inp)*cos(y_inp);
    input->output[0][2] = sin(y_inp);

    // clear map activity
    map->reset_output();

    // activate and settle
    for (int k=0; k<15; ++k) {
	if (k==0)
		map->activate(1);
	else 
		map->activate(0);

	// display
        if (i%disp_rate==0) {

		// calculate c.o.g.
		calc_cog(cog,map,0);
	}
    }

    // learn
    map->learn();
    map->normalize_weights();

    // update sigmoid
    if (i<1300) {
      	map->sig_low += (map->sig_high-map->sig_low)/2.0/(steps/20.0);
     	// cout << "new sig_low = " << map->sig_low << endl; 
      	//map->sig_high -= (map->sig_high-map->sig_low)/4.0/(steps/10.0);
    }

    // reduce learing rate
    /*
    if (i%600==0 && i!=0) { 
	if (map->alpha_affs[0]>0.001) {
      	REAL new_alpha = map->alpha_affs[0];
      	map->alpha_affs[0] = new_alpha * 0.1;
      	map->alpha_lat_exc = map->alpha_lat_exc * 0.1 ;
      	map->alpha_lat_inh = map->alpha_lat_inh * 0.1 ;
       	cout << "new alpha = " << new_alpha << endl;     
	}
    }
    */
	
    // reduce radius
    if (i==300) {
	cout << "lissom.cc:run():  reducing radius.\n";
	netlog->append("iter %d: lissom.cc:run():  reducing radius to %d\n",
		i, 2);
	map->resize_lat_exc_rad(2);
    }
    if (i==600) {
	cout << "lissom.cc:run():  reducing radius.\n";
	netlog->append("iter %d: lissom.cc:run():  reducing radius to %d\n",
		i, 1);
	map->resize_lat_exc_rad(1);
    }

    // kill connections
    if (i%1000==0 && i!=0 && i<3000) {
	cout << "lissom.cc:run():  killing lat inh weights.";
	netlog->append("iter %d: lissom.cc:run():  killing lat inh wgts.",i);
	map->kill_lat_weights(0.001);
    }

    // log every 1000 iterations
    if (i%disp_rate==0)
  	netlog->append("iter %d reached.",i);

  }

  // set start of next run
  t = i;

}


//----------------------------------------------------------------------------
//	save() 
//	  - save blackboard
//----------------------------------------------------------------------------
void save() {

  cout << "lissom.cc: saveit() : saving to lissom.sav\n";
  map->net_bb->save("lissom.sav");

}
//----------------------------------------------------------------------------
//	main() 
//	  - the MAIN
//----------------------------------------------------------------------------
int main(int argc, char *argv[]) {

  steps = 20000;
  t 	= 0;

  // check args
  if (argc < 2) {
	cout << "usage: " << argv[0] << " <configfilename> [<network imagefile>] [<start t>]\n";
 	exit(0);
  } 

  // setup network
  setup(argv[1]);

  // if network imagefilename given, load the network image
  if (argc >= 3) {
	cout << "lissom.cc : main() : loading image " << argv[2] << "\n";
	map->net_bb->load(argv[2]);
  }

  if (argc >= 4) {
	cout << "lissom.cc : main() : setting t = " << argv[3] << "\n";
	t = atoi(argv[3]);
  }

  // setup the windows
  run();
  save();

  delete netlog;

}
