#!/usr/bin/perl

$usage = "analyze-or-cog <orfile> <cogfile> <rowbegin> <rowend> <colbegin> <colend> <distance>\n";

if ($#ARGV != 6) {
	print $usage;
	exit;
}

$orfile = $ARGV[0];
$cogfile = $ARGV[1];
$rowbegin = $ARGV[2];
$rowend = $ARGV[3];
$colbegin = $ARGV[4];
$colend = $ARGV[5];
$dist = $ARGV[6];

# read orfile
open(FP,$orfile);
$k=0;
while (<FP>) {
  chop;
  @tmprow = split(/[\s]+/);
  for ($i=0; $i<= $#tmprow; ++$i) {
	$ormat[$k][$i] = $tmprow[$i];
  }
  $k++;
}
close(FP);

$cols = $#tmprow;

# read cogfile
open(FP,$cogfile);
$k=0;
while (<FP>) {
  chop;
  @tmprow = split(/[\s]+/);
  for ($i=0; $i<= $#tmprow; ++$i) {
	($x,$y) = split(/:/,$tmprow[$i]);
	$cogx[$k][$i] = $x;
	$cogy[$k][$i] = $y;
  }
  $k++;
}
close(FP);

# go through given row and print out or-diff vs. distance
$step = 1;
for ($row=$rowbegin; $row<=$rowend; $row+=$step) {

	for ($i=$colbegin; $i<=$colend; $i+=$step) {

	  # or gradient;
	  $ordiff = &calc_ordiff($ormat[$row][$i],$ormat[$row][$i+$dist]);

	  # $ordiff = $ormat[$row][$i]-$ormat[$row][$i+$dist];
	  $d   = sqrt(($cogx[$row][$i]-$cogx[$row][$i+$dist])**2+
			 ($cogy[$row][$i]-$cogy[$row][$i+$dist])**2);
	  print "$ordiff $d\n";
	}
}
  

sub min {

  my ($x,$y) = @_;

  if ($x>$y) {
	return $y;
  } else {
	return $x;
  }
}

sub max {

  my ($x,$y) = @_;

  if ($x>$y) {
	return $x;
  } else {
	return $y;
  }
}

sub calc_ordiff {
  my ($or1,$or2) = @_;

  my $mx = &max($or1,$or2);
  my $mn = &min($or1,$or2);

  my $ordiff = $mx-$mn;
  if ($ordiff > 0.5) {
     $ordiff = 1.0-$ordiff;
  }
  $ordiff *=180;
  return $ordiff;

}

